/*
Copyright (c) 2019 zoranc on pokemon-mini.net forum

Permission to use, copy, modify, and distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include "images.h"

#include "pm.h"
#include "small_char_font.h"

#include <stdio.h>
#include <string.h>

typedef unsigned char byte;

extern const byte levels_tile_index[75][9][13];

volatile byte flag = 0;

#define SPRITE_SIZE 64
#define TITLE_START 0
#define TITLE_LEN 3
#define TITLE_LONG_LEN 10
#define BALL_SPRITE (TITLE_START + TITLE_LONG_LEN)
#define LINE_START (BALL_SPRITE + 2)
#define LINE_LEN 6
#define ALL_SPRITES (LINE_START + LINE_LEN)

#define BEGIN_FLIP 16
#define END_FLIP 24
#define BEGIN_ROT 24
#define END_ROT 28
#define TILE_OFFSET_X (16 - 8)
#define TILE_OFFSET_Y (16 - 8)

#define LEVEL_NUMBER 75

// box interaction states
#define ENTER   0
#define IN      1
#define EXIT    2

// direction types
#define DIR_UP      0
#define DIR_RIGHT   1
#define DIR_DOWN    2
#define DIR_LEFT    3
#define DIR_STOP    4
#define DIR_MASK    0x03

// tile types
#define EMPTY       0
#define SEED        1
#define TELEPORT1   2
#define TELEPORT2   3
#define STONE       4
#define ALONE       5
#define HORIZONTAL  6
#define VERTICAL    7
#define WUL         8
#define WUR         9
#define WDR         10
#define WDL         11
#define WTU         12
#define WTR         13
#define WTD         14
#define WTL         15
#define EMPTY_UL    16
#define EMPTY_UR    17
#define EMPTY_DR    18
#define EMPTY_DL    19
#define SHADOW_UL   20
#define SHADOW_UR   21
#define SHADOW_DR   22
#define SHADOW_DL   23
#define ROT_UL      24
#define ROT_UR      25
#define ROT_DR      26
#define ROT_DL      27
#define SOLID_UL    28
#define SOLID_UR    29
#define SOLID_DR    30
#define SOLID_DL    31
#define POWER_U     32
#define POWER_R     33
#define POWER_D     34
#define POWER_L     35
#define POWER_TU    36
#define POWER_TR    37
#define POWER_TD    38
#define POWER_TL    39
#define PLAYER_U    40
#define PLAYER_R    41
#define PLAYER_D    42
#define PLAYER_L    43

// tile types
#define TYPE_WALL       0x00
#define TYPE_TRIANGLE   0x10
#define TYPE_POWER      0x20
#define TYPE_MASK       0xf0

// triangle tile types
#define TRG_EMPTY      0x10
#define TRG_SHADOW     0x14
#define TRG_ROT        0x18
#define TRG_SOLID      0x1c
#define TRG_MASK       0xfc

// triangle tile orientations
#define ORIENT_UL       0
#define ORIENT_UR       1
#define ORIENT_DR       2
#define ORIENT_DL       3
#define ORIENT_MASK     0x03


byte num_flip;
byte index_flip[50];
byte num_rot;
byte index_rot[50];
byte num_seed;

signed char dx = 0, dy = 0;
byte x, y;
byte bx, by;
byte old_bx, old_by;
byte px, py;
byte tile, last_tile;
byte box_state = EXIT;
byte direction = DIR_UP;
byte tile_offset_x = TILE_OFFSET_X, tile_offset_y=TILE_OFFSET_Y;
byte tele_coord[2][2][2];
byte teleports[2];

#define STATE_NONE 0
#define STATE_START 1
#define STATE_EXPLANATION 2
#define STATE_TITLE 3
#define STATE_LOAD 4
#define STATE_PLAY 5
#define STATE_FINISHED 6
#define STATE_GAME_OVER 7

byte game_state;
int level;

byte shade = 0;

#define MIN_PRESS 1
#define MAX_PRESS 15
#define BALL_SPEED 2
#define DECREASE_FRAME_CNT() --frame_cnt; --frame_cnt
//#define BALL_SPEED 3
//#define DECREASE_FRAME_CNT() --frame_cnt; --frame_cnt; --frame_cnt
byte a_cnt;
byte a_pressed;
byte b_cnt;
byte b_pressed;
byte up_cnt;
byte up_pressed;
byte down_cnt;
byte down_pressed;
byte frame_cnt;
byte blinking_title;

byte draw_line;
byte exist_line;
byte line_sprites;
unsigned line_size;
byte line_horizontal;
unsigned sprite_mask, sprite_draw;
byte lx1, ly1, lx2, ly2;

byte sprites[ALL_SPRITES * SPRITE_SIZE] _at(0x1500);

const byte ball_sprite[] = {
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // mask UL
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, // mask DL
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // draw UL
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // draw DL
        0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // mask UR
        0xfc, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // mask DR
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // draw UR
        0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // draw DR

        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01,
        0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
        0xfc, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
        0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x03, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
};

const byte is_head_on[]  = {
// triangle types / directions
// UL, UR, DR, DL
    1,  1,  0,  0,  // UP
    0,  1,  1,  0,  // RIGHT
    0,  0,  1,  1,  // DOWN
    1,  0,  0,  1,  // LEFT
};

const byte next_direction[]  = {
    // triangle types                               /  directions
    // UL,      UR,         DR,         DL
    DIR_STOP,   DIR_STOP,   DIR_RIGHT,  DIR_LEFT,   // UP
    DIR_UP,     DIR_STOP,   DIR_STOP,   DIR_DOWN,   // RIGHT
    DIR_LEFT,   DIR_RIGHT,  DIR_STOP,   DIR_STOP,   // DOWN
    DIR_STOP,   DIR_UP,     DIR_DOWN,   DIR_STOP,   // LEFT
};

// forward declarations
void cancel_line(void);

void change_direction(byte dir) {
    direction = dir;
    if(draw_line)
        cancel_line();
    switch (direction) {
    case DIR_UP:
        dx = 0;
        dy = -1;
        break;
    case DIR_RIGHT:
        dx = 1;
        dy = 0;
        break;
    case DIR_DOWN:
        dx = 0;
        dy = 1;
        break;
    case DIR_LEFT:
        dx = -1;
        dy = 0;
        break;
    case DIR_STOP:
        dx = 0;
        dy = 0;
        break;
    }
}


void set_offsets(int x, int y) {
    PRC_SCROLL_X = x;
    PRC_SCROLL_Y = y;
    tile_offset_x = -x + TILE_OFFSET_X;
    tile_offset_y = -y + TILE_OFFSET_Y;
}

void clear_level(void) {
    byte c, r;

    num_flip = 0;
    num_rot = 0;
    num_seed = 1;
    x = y = 0;
    teleports[0] = teleports[1] = 0;
    for(r=0; r<10; r++)
      for(c=0; c<14; c++)
          TILEMAP[r*16+c] = EMPTY;
    change_direction(DIR_STOP);
    OAM[BALL_SPRITE].ctrl = 0;
}

void copy_level(const byte level_data[9][13])
{
    byte c, r;
    byte tile;
    byte index;
    byte telep;

    num_flip = 0;
    num_rot = 0;
    num_seed = 0;
    teleports[0] = teleports[1] = 0;

    x = y = 4;
    for(c=0; c<13; c++)
        if(level_data[8][c] != EMPTY) { y += 4; break; }
    for(c=0; c<13; c++)
        if(level_data[0][c] != EMPTY) { y -= 4; break; }
    for(r=0; r<9; r++)
        if(level_data[r][12] != EMPTY) { x += 4; break; }
    for(r=0; r<9; r++)
        if(level_data[r][0] != EMPTY) { x -= 4; break; }
    set_offsets(x, y);

    for(r=0; r<9; r++)
      for(c=0; c<13; c++)
      {
        tile = level_data[r][c];
        index = r*16+c;

        if(tile == SEED)
            ++num_seed;
        if(BEGIN_FLIP <= tile && tile < END_FLIP)
            index_flip[num_flip++] = index;
        else if(BEGIN_ROT <= tile && tile < END_ROT)
            index_rot[num_rot++] = index;
        else if(PLAYER_U <= tile && tile <= PLAYER_L) {
            x = c;
            y = r;
            x = x * 8 + 4;
            y = y * 8 + 4;
            change_direction(tile & DIR_MASK);
            tile = EMPTY;
        } else if(tile == TELEPORT1 || tile == TELEPORT2) {
            telep = tile == TELEPORT1 ? 0 : 1;
            tele_coord[telep][teleports[telep]][0] = c;
            tele_coord[telep][teleports[telep]][1] = r;
            ++teleports[telep];
        }

        TILEMAP[index] = tile;
      }
}

_interrupt(2) void prc_frame_copy_irq(void)
{
    flag = 1;
    ++frame_cnt;
    IRQ_ACT1 = IRQ1_PRC_COMPLETE;
}

byte print_title_char(byte pos, byte row, unsigned char ch)
{
    byte i;
    const unsigned char _rom *p2;
    int pos1 = pos >> 3, pos2 = pos & 0x7;
    byte *p = sprites + TITLE_START * SPRITE_SIZE + 16 + (pos1 << 5) + pos2 + (row << 3);
    unsigned char len;

    if(ch < 0x20)
        ch = 0;
    else
        ch -= 0x20;
    if(ch>=0x80)
        ch = 0;
    p2 = small_char_font[ch];
    len = *p2++;
    for(i = 0; i<len; i++)
    {
        *p++ = *p2++;
        if((((long)p) & 0x7) == 0) {
            p += 24;
        }
    }
    return pos + len;
}
struct Row {
    byte pos;
    const char *str;
};

const struct Row row[8][4] = {
    {
        { 15, "This game is" },
        { 9, "based on game" },
        { 90, "called 'Pazuru'."},
        { 104, "Enjoy it!"}
    },
    {
        { 21, "Press (A)" },
        { 20, "to rotate" },
        { 94, "the bumpers" },
        { 91, "with a border."},
    },
    {
        { 17, "Spikes are" },
        { 16, "dangerous!" },
        { 81, "Try to avoid them."},
        { 0, NULL}
    },
    {
        { 10, "Some bumpers" },
        { 6, "turn on and off" },
        { 100, "when you" },
        { 100, "press (A)."},
    },
    {
        { 15, "The stones" },
        { 21, "can take" },
        { 91, "one more hit!" },
        { 0, NULL },
    },
    {
        { 6, "The crosses will" },
        { 13, "teleport you" },
        { 82, "to the other one." },
        { 0, NULL },
    },
    {
        { 1, "Press and hold (B)" },
        { 8, "to draw a wall." },
        { 82, "Use it as a barier" },
        { 88, "against spikes!" },
    },
    {
        { 9, "Congatulations!" },
        { 0, "" },
        { 89, "You have beat" },
        { 100, "the game!"}
    }
};

int explanation_number(byte level) {
    switch (level) {
    case 1: return 1;
    case 2: return 2;
    case 9: return 3;
    case 13: return 4;
    case 23: return 5;
    case 31: return 6;
    default: return -1;
    }
}

#define END_TILE_LEN 10
#define HALF_TITLE_LEN 5

void create_big_title(byte num) {
    byte i;
    const char *p;
    byte pos;
    byte rows = row[num][3].str ? 4 : 3;
    memset(sprites, 0, END_TILE_LEN * SPRITE_SIZE);
    for(i = 0; i < rows; ++i)
        for(pos = row[num][i].pos, p = row[num][i].str; *p; ++p)
            pos = print_title_char(pos, i & 1, *p);
    rows = row[num][3].str ? 32 : 36;
    for(i = 0; i < END_TILE_LEN; ++i) {
        OAM[i].x = (i%HALF_TITLE_LEN) * 16 + 24;
        OAM[i].y = (i/HALF_TITLE_LEN) * 16 + rows;
        OAM[i].tile = i;
        OAM[i].ctrl = OAM_ENABLE;
    }
}

void create_end_title(void) {
    create_big_title(sizeof(row) / sizeof(row[0]) - 1);
}

void create_title(void) {
    byte i;
    char *p;
    byte pos;
    byte num = level + 1;

    memset(sprites + TITLE_START * SPRITE_SIZE, 0, TITLE_LEN * SPRITE_SIZE);

    for(pos = 3, p = "Level - "; *p; ++p) pos = print_title_char(pos, 0, *p);
    if(num>9) pos = print_title_char(pos, 0, (num / 10) + '0');
    pos = print_title_char(pos, 0, (num % 10) + '0');
    for(pos = 3, p = "Press (A)"; *p; ++p) pos = print_title_char(pos, 1, *p);

    for(i = TITLE_START; i < TITLE_START + TITLE_LEN; ++i) {
        OAM[i].x = i * 16 + 40;
        OAM[i].y = 40;
        OAM[i].tile = i;
        OAM[i].ctrl = OAM_ENABLE;
    }
    blinking_title = 0;
}

void enable_title(void) {
    byte i;
    for(i = TITLE_START; i < TITLE_START + TITLE_LEN; ++i) {
        OAM[i].ctrl = OAM_ENABLE;
    }
}

void disable_title(void) {
    byte i;
    for(i = TITLE_START; i < TITLE_START + TITLE_LONG_LEN; ++i) {
        OAM[i].ctrl = 0;
    }
}

void start_line(void) {
    byte i;
    byte *pos = sprites + LINE_START * SPRITE_SIZE;
    for (i=0; i<LINE_LEN; ++i) {
        memset(pos, 0xff, 16); pos += 16;
        memset(pos, 0x00, 16); pos += 16;
        memset(pos, 0xff, 16); pos += 16;
        memset(pos, 0x00, 16); pos += 16;
    }
    for (i=0; i<LINE_LEN; ++i) {
        OAM[i + LINE_START].tile = i + LINE_START;
    }
    line_size = 0;
    line_horizontal = (dy==0);
    line_sprites = 1;
    sprite_mask = LINE_START * SPRITE_SIZE;
    sprite_draw = sprite_mask + 16;
    sprites[sprite_mask] = 0xfe;
    sprites[sprite_draw] = 0x01;
    OAM[LINE_START].x = x + tile_offset_x + 8;
    OAM[LINE_START].y = y + tile_offset_y + 8;
    OAM[LINE_START].ctrl = OAM_ENABLE;
    draw_line = 1;
    exist_line = 0;
}

void finish_line(void) {
    lx1 = lx2 = OAM[LINE_START].x - (tile_offset_x + 8);
    ly1 = ly2 = OAM[LINE_START].y - (tile_offset_y + 8);
    if (line_horizontal) lx2 += line_size;
    else ly2 += line_size;
    draw_line = 0;
    exist_line = 1;
}

void continue_line(void) {
    byte i;
    ++line_size;
    switch (direction) {
    case DIR_LEFT:
        for(i=0; i<line_sprites; ++i)
            --OAM[LINE_START + i].x;
    case DIR_RIGHT:
        ++sprite_mask;
        ++sprite_draw;
        if(((line_size) & 0x07) == 0) {
            sprite_mask += 24; // 64 / 2 - 8
            sprite_draw += 24;
        }
        if(((line_size) & 0x0f) == 0) {
            OAM[LINE_START + line_sprites].x = OAM[LINE_START + line_sprites - 1].x + 16;
            OAM[LINE_START + line_sprites].y = OAM[LINE_START + line_sprites - 1].y;
            OAM[LINE_START + line_sprites].ctrl = OAM_ENABLE;
            ++line_sprites;
        }
        sprites[sprite_mask] = 0xfe;
        sprites[sprite_draw] = 0x01;
        break;
    case DIR_UP:
        for(i=0; i<line_sprites; ++i)
            --OAM[LINE_START + i].y;
    case DIR_DOWN:
        if(((line_size) & 0x07) == 0x0) {
            sprite_mask += 8;
            sprite_draw += 8;
        }
        if(((line_size) & 0x0f) == 0x0) {
            sprite_mask += 48; // 64 - 2 * 8
            sprite_draw += 48;
            OAM[LINE_START + line_sprites].x = OAM[LINE_START + line_sprites - 1].x;
            OAM[LINE_START + line_sprites].y = OAM[LINE_START + line_sprites - 1].y + 16;
            OAM[LINE_START + line_sprites].ctrl = OAM_ENABLE;
            ++line_sprites;
        }
        sprites[sprite_mask] <<= 1;
        sprites[sprite_draw] <<= 1;
        ++sprites[sprite_draw];
        break;
    case DIR_STOP:
        break;
    }
}

void cancel_line(void) {
    byte i;
    for(i = 0; i < line_sprites; ++i)
        OAM[i + LINE_START].ctrl = 0;
    draw_line = 0;
    exist_line = 0;
}

void check_line_hit(void) {
    byte nx = x + dx;
    byte ny = y + dy;
    if (nx != lx1 && ny != ly1)
        return;
    if (lx1 <= nx && nx <= lx2 && ly1 <= ny && ny <= ly2) {
        if (line_horizontal != (dy==0))
            change_direction(direction ^ 2);
        cancel_line();
    }
}

void pressed_a(void)
{
    byte v;
    byte i;

    for(i = 0; i < num_flip; ++i)
    {
        TILEMAP[index_flip[i]] ^= 0x04;
    }
    for(i = 0; i < num_rot; ++i)
    {
        v = TILEMAP[index_rot[i]] + 1;
        if(!(v&0x03)) v -= 0x04;
        TILEMAP[index_rot[i]] = v;
    }
}

void pressed_b(void)
{
    if (exist_line)
        cancel_line();
    if (!draw_line && tile == EMPTY)
        start_line();
}

void released_b(void)
{
    if (draw_line)
        finish_line();
}

void hit_stone(void) {
    TILEMAP[by*16+bx] = EMPTY;
    change_direction(direction ^ 0x2);
    box_state = EXIT;
}

void handle_triangle(void) {
    byte type = tile & TRG_MASK;
    byte orient = tile & ORIENT_MASK;
    if (px==4 && py==4) {
        if (type == TRG_EMPTY)
            box_state = EXIT;
        else {
            if (is_head_on[(direction & 0x03) << 2 | orient]) {
                if (type == TRG_SHADOW) {
                    box_state = EXIT;
                    return;
                }
                orient = last_tile & ORIENT_MASK;
            }
            change_direction(next_direction[(direction & 0x03) << 2 | orient]);
            box_state = EXIT;
        }
    } else {
        if (type != TRG_EMPTY) {
            if (is_head_on[(direction & 0x03) << 2 | orient]) {
                change_direction(direction ^ 0x2);
                box_state = EXIT;
            } else {
                last_tile = tile;
            }
        }
    }
}

void process_enter(void) {
    byte telep;

    switch (tile & TYPE_MASK) {
    case TYPE_POWER:
        change_direction(DIR_STOP); // DEAD!
        game_state = STATE_FINISHED;
        box_state = EXIT;
        break;
    case TYPE_TRIANGLE:
        if (draw_line && tile >= SHADOW_UL) finish_line();
        if (px!=0 && py!=0) {
            box_state = IN;
            handle_triangle();
        }
        break;
    case TYPE_WALL:
        if (draw_line && tile > SEED) finish_line();
        switch (tile) {
        case EMPTY:
            box_state = EXIT;
            break;
        case SEED:
            if (px==4 && py==4) {
                TILEMAP[by*16+bx] = EMPTY;
                --num_seed;
                if(num_seed==0) {
                    change_direction(DIR_STOP); // DONE!
                    game_state = STATE_FINISHED;
                }
                box_state = EXIT;
            }
            break;
        case STONE:
            switch (direction) {
            case DIR_UP: if (py==6) hit_stone(); break;
            case DIR_LEFT: if (px==6) hit_stone(); break;
            case DIR_DOWN: if (py==2) hit_stone(); break;
            case DIR_RIGHT: if (px==2) hit_stone(); break;
            }
            break;
        case TELEPORT1:
        case TELEPORT2:
            if (px==4 && py==4) {
                telep = (tile == TELEPORT1) ? 0 : 1;
                if(tele_coord[telep][0][0] == bx && tele_coord[telep][0][1] == by) {
                    x = tele_coord[telep][1][0];
                    y = tele_coord[telep][1][1];
                } else {
                    x = tele_coord[telep][0][0];
                    y = tele_coord[telep][0][1];
                }
                x = x * 8 + 4;
                y = y * 8 + 4;
                box_state = EXIT;
            }
            break;
        default:
            // all walls here
            if (py==2 || py==5 || px==2 || px==5) {
                change_direction(direction ^ 0x2);
                box_state = EXIT;
            }
            break;
        }
    }
}

void process_in(void) {
    handle_triangle();
}

void process_exit(void) {
}

void advance_point(void) {
    x += dx;
    y += dy;
    // where in the tile - each tile is 8x8
    old_bx = bx;
    old_by = by;
    bx = x >> 3;
    by = y >> 3;
    px = x & 7;
    py = y & 7;
    if(bx != old_bx || by != old_by)
        box_state = ENTER;
    tile = TILEMAP[by*16+bx];
    if(draw_line) {
        if(tile==EMPTY)
            continue_line();
        else
            cancel_line();
    } else if(exist_line) {
        check_line_hit();
    }
}

void process_state(void)
{
    switch(box_state) {
        case ENTER: process_enter(); break;
        case IN: process_in(); break;
        case EXIT: process_exit(); break;
    }
}

void prc_copy(void)
{
    flag = 0;

    if(game_state == STATE_PLAY) {
        while(frame_cnt > BALL_SPEED) {
            DECREASE_FRAME_CNT();
            advance_point();
        }

        OAM[BALL_SPRITE].x = x + tile_offset_x;
        OAM[BALL_SPRITE].y = y + tile_offset_y;

        if(!(KEY_PAD&KEY_A)) {
            if(a_cnt < MAX_PRESS) ++a_cnt;
        } else if(a_cnt) {
            if(a_cnt >= MIN_PRESS && a_cnt < MAX_PRESS) a_pressed = 1;
            a_cnt = 0;
        }
        if(a_pressed) {
            // ROTATE ACTION - short A
            a_pressed = 0;
            pressed_a();
        }
        if(a_cnt == MAX_PRESS || !(KEY_PAD&KEY_B)) {
            // HOLD ACTION - long A or short/long  B
            pressed_b();
        }
        if(a_cnt==0 && (KEY_PAD&KEY_B)) {
            released_b();
        }

        process_state();

    } else if(game_state == STATE_TITLE) {
        if(!(KEY_PAD&KEY_A)) {
            if(a_cnt < MIN_PRESS) ++a_cnt;
        } else {
            if(a_cnt >= MIN_PRESS) a_pressed = 1;
            a_cnt = 0;
        }
        if(!(KEY_PAD&KEY_UP)) {
            if(up_cnt < MIN_PRESS) ++up_cnt;
        } else {
            if(up_cnt >= MIN_PRESS) up_pressed = 1;
            up_cnt = 0;
        }
        if(!(KEY_PAD&KEY_DOWN)) {
            if(down_cnt < MIN_PRESS) ++down_cnt;
        } else {
            if(down_cnt >= MIN_PRESS) down_pressed = 1;
            down_cnt = 0;
        }
        if(a_pressed) {
            a_pressed = 0;
            game_state = STATE_PLAY;
        }
        if(up_pressed) {
            up_pressed = 0;
            if (level > 0) {
                --level;
                game_state = STATE_LOAD;
            }
        }
        if(down_pressed) {
            down_pressed = 0;
            if (level < LEVEL_NUMBER - 1) {
                ++level;
                game_state = STATE_LOAD;
            }
        }
        ++blinking_title;
        if(blinking_title==15) disable_title();
        if(blinking_title==25) { enable_title(); blinking_title=0; }
    } else if (game_state == STATE_START || game_state == STATE_EXPLANATION) {
        if(!(KEY_PAD&KEY_A)) {
            if(a_cnt < MIN_PRESS) ++a_cnt;
        } else {
            if(a_cnt >= MIN_PRESS) a_pressed = 1;
            a_cnt = 0;
        }
        if(a_pressed) {
            a_pressed = 0;
            game_state = STATE_NONE;
        }
        return;
    }

    shade ^= 1;
    if(shade) {
        PRC_MAP = images_frame0;
        OAM[BALL_SPRITE].tile = BALL_SPRITE;
    } else {
        PRC_MAP = images_frame1;
        OAM[BALL_SPRITE].tile = BALL_SPRITE + 1;
    }
}

int main(void)
{
    PRC_MODE = COPY_ENABLE|SPRITE_ENABLE|MAP_ENABLE|MAP_16X12;
    PRC_RATE = RATE_36FPS;

    set_offsets(4, 4);

    PRC_MAP = images_frame0;

    PRC_SPR = sprites;

    game_state = STATE_START;

    // Enable PRC IRQ
    IRQ_ENA1 = IRQ1_PRC_COMPLETE;
//    PRI_PRC(1);
    IRQ_PRI1 = 0x40;

    while(1) {
        frame_cnt = 0;
        while(game_state == STATE_PLAY) {
            if(flag) {
                prc_copy();
            }
        }

        cancel_line();

        if(game_state == STATE_START) {
            create_big_title(0);
            while (game_state == STATE_START) {
                if(flag) {
                    prc_copy();
                }
            }
            disable_title();
            num_seed = 1;
            level = 0;
            game_state = STATE_FINISHED;
        }

        if(game_state == STATE_FINISHED) {
            if (num_seed == 0) ++level;
            if (level == LEVEL_NUMBER) {
                clear_level();
                create_end_title();
                game_state = STATE_GAME_OVER;
            } else if(explanation_number(level + 1) >= 0) {
                clear_level();
                create_big_title(explanation_number(level + 1));
                game_state = STATE_EXPLANATION;
            } else {
                game_state = STATE_LOAD;
            }
        }

        if(game_state == STATE_EXPLANATION) {
            while(game_state == STATE_EXPLANATION) {
                if(flag) {
                    prc_copy();
                }
            }
            disable_title();
            memcpy(sprites + SPRITE_SIZE * BALL_SPRITE, ball_sprite, 2 * SPRITE_SIZE);
            game_state = STATE_LOAD;
        }

        if(game_state == STATE_LOAD) {
            copy_level(levels_tile_index[level]);

            OAM[BALL_SPRITE].x = x + tile_offset_x;
            OAM[BALL_SPRITE].y = y + tile_offset_y;
            OAM[BALL_SPRITE].tile = BALL_SPRITE;
            OAM[BALL_SPRITE].ctrl = OAM_ENABLE;

            shade = 0;
            flag = 0;
            a_cnt = 0;
            a_pressed = 0;
            b_cnt = 0;
            b_pressed = 0;
            create_title();
            game_state = STATE_TITLE;
        }

        while(game_state == STATE_GAME_OVER) {
            // nothing
        }

        while(game_state == STATE_TITLE) {
            if(flag) {
                prc_copy();
            }
        }
        disable_title();
    }
}

